/*************************************************************
 * tweetGeo                                       (8 Sep 2013)
 * www.jsylvest.com/tweetmap/
 * 
 * Jared Sylvester
 * jared@jsylvest.com
 * www.cs.umd.edu/~jared/
 * www.jsylvest.com
 * 
 * Created for the Center for Complexity in Business
 *   at the University of Maryland College Park
 *   http://www.rhsmith.umd.edu/ccb/
 *
 * Released under GPL v3 License (see license.txt)
 *   Feel free to use and modify this code as you wish
 *   but be a mensch and throw a little credit my way
 *
 * This code is very much still a work in progress.
 *   I take no liability, implicit or otherwise, etc. etc.
 *************************************************************/


/* the following libraries are used to convert UNIX timestamps
 to human-readable formats */
import org.joda.time.format.*;
import org.joda.time.convert.*;
import org.joda.time.*;
import org.joda.time.tz.*;
import org.joda.time.chrono.*;
import org.joda.time.base.*;
import org.joda.time.field.*;
import java.text.*; 
import java.util.*; 

/* two modes:
 1. "cumulative" mode:
 Every tweet which has happened so far is shown on the map.
 2. "instantaneous" (or "fade") mode:
 Tweets are rendered when they occur and then fade out.
 The amount of time they're visible is determined by windowSize*/
boolean cumulativeMode = false;

/* save each frame so you can render a video? */
boolean saveFrames = true;

/* what it says on the tin */
boolean showCoastline =   true;
boolean showIntlBorders = true;
boolean showStormEye =    true;
boolean showStormTrack =  true;
boolean showTweets =      true;
boolean showDonations =   true;

/* input data files; see documentation */
String tweetFile =          "sandy_tweets.csv";
String donationFile =       "sandy_donations.csv";
String eyeFile =            "sandy_eye.csv";
String trackFile =          "sandy_track.csv";

/* name of map as it appears on screen */
String datasetDisplayName = "Hurricane Sandy";

/* prefix for output file names */
String outputName =         "sandy_kavray_glb";

/* various map projections are available; see documentation */
final int WINKEL          = 10001;
final int KAVRAY          = 10002;
final int ALBERS_CON_EA   = 10003;
final int LAMBERT_CYL_EA  = 10004;
final int GALL_STEREO     = 10005;
final int MERCATOR        = 10006;
int projectionType = KAVRAY;
String projName = "";  // set in setDisplayName()

/* how zoomed-in are we going to be? */
final int SCALE_GLOBE     = 20001;
final int SCALE_NORTHWEST = 20002;
final int SCALE_LOWER48   = 20003;
final int SCALE_FREE      = 20004;
int projectionScale = SCALE_GLOBE;
/* scale_free is a new feature I'm adding to allow 
 panning and zooming. It currently only works
 (to the extent it does) with Winkel & Kavrayskiy */

/* number of tweets */
int n = 0;
/* the time of each tweet */
float[] times;
/* the latitude and longitude of each tweet */
float[] lat;
float[] lon;

/* The following is used for the twitter account which spoofed
 the hurricane eye's location. It's displayed as either a
 red circle or blue dot, depending on mode */
float[] pathTimes;
float[] pathLat;
float[] pathLon;

/* this is the path of the storm
 it's displayed on the map as a wide gray line */
PVector[] track;

/* this stores the data on mobile donations from the Red Cross */
/*   it's a very simple class defined in redcross.pde */
Donation[] donations;

/* this is the current time (in secs) within the playback */
float t = 0;

/* stepSize is the number of seconds to move the clock
 ahead each frame of the animation */
float stepSize = 288.0;
/* Letting stepSize==288 will compress 1day of data to 10sec
 in the animation. To compress 1 day into T seconds, set
 stepSize = (24*60*60)/(r*T) where r is the framerate
 you're rendering out at */

/* windowSize is the number of seconds a tweet will appear
 for in instant mode */
float windowSize = stepSize*20;

/* the following will be set in readTweets() */
double minTime = 1999999999;
double maxTime = 0;
double duration = 1;

/* for rendering text to screen */
PFont fontSrcCode;

/* set in readCoasts()
 (will be lengthened as needed when reading) */
PShape[] coasts = new PShape[128];
int nCoasts = -1;

/* set in readBorders()
 (will be lengthened as needed when reading) */
PShape[] borders = new PShape[128];
int nBorders = -1;

/* Used to coarsen the border and coasts data
 any point which is within eps degrees lat/lon of the previous
 is ignored to avoid extraneous detail.
 Smaller values will result in more fine-grained curves */
float epsilon = 0.01; 

/* how much space to leave around map edges (in some projections only) */
float margin = 5.0;

/* These  are used in the scale_free mode to control the view */
float longioffset = 0;
float latioffset = 0;
float scaleFactor = 1.0;

/* set this in setup(), since window since must be defined already */
float rescale;

/* pre-calculate this to use w/ Kavrayskiy proj */
final float twoOvrSqrt3 = 2.0/sqrt(3.0);

/*** *** *** *** SETUP *** *** *** ***/
void setup() {
  size( 1280, 720, P2D );
  frameRate(30);
  smooth(8);
  ellipseMode(RADIUS);

  /*
  Ignore this business about loading properties. I just threw
  it in at the last minute to facilitate rendering out a big
  batch of files at the last minute. It only works when you export
  to a stand-alone Application, and it's a big pain. You can
  safely ignore it.
  Properties theProps = loadCommandLine();
  evalProps(theProps);
  println(theProps.toString());
  */  
  
  fontSrcCode = createFont("SourceCodePro-Regular", 20, true);
  textFont(fontSrcCode);
  /*
  If you do not have the Source Code Pro font, you can get it free from Adobe or GitHub
   http://store1.adobe.com/cfusion/store/html/index.cfm?store=OLS-US&event=displayFontPackage&code=1960
   https://github.com/adobe/source-code-pro
   Alternately you can choose another font.
   To see which fonts are available to Processing on your
   system, run the following line in a sketch:
   println(PFont.list());
   */

  setDisplayName();

  rescale = (width*(height-2*margin)) / (height*(width-2*margin));

  /* read all the necessary data */
  readTweets(tweetFile);
  if ( showStormEye ) {
    readEye(eyeFile);
  }
  if ( showStormTrack ) {
    readTrack(trackFile);
  }
  if ( showCoastline ) {
    readCoasts("noaa_coast.dat");
  } 
  if ( showIntlBorders ) { 
    readBorders("intl_borders.dat");
  }
  if ( showDonations ) {
    readDonations(donationFile);
  }

  if ( projectionScale == SCALE_NORTHWEST ) {
    longioffset = 90.0;
  } 
  else if ( projectionScale == SCALE_LOWER48 ) {
    longioffset = 96.0;
  }

  /*
  // for debugging map projections
   println( "X: "+getX(-90,-180)+", "+getX(90,180) );
   println( "Y: "+getY(-90,0)+", "+getY(90, -180) );
   println( "UL: ( "+getX(52,-131)+", "+getY(52,-131)+" )");
   println( "LR: ( "+getX(23,-65)+", "+getY(23,-65)+" )");
   */
  /*
  // for debugging time stamps
   long temp = int(minTime);
   String date = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss").format(new java.util.Date(temp*1000));
   println(date);
   Date foo = new java.util.Date(temp*1000);
   println(foo.getHours()+":"+nf(foo.getMinutes(),2)+"."+foo.getSeconds());
   drawSun(foo);
   */
}

/*** *** *** *** DRAW *** *** *** ***/
void draw() {
  background(219, 219, 219); // Gainsboro gray

  /* change the size of dots as maps scale changes */
  float r = 2.75;
  if ( projectionScale == SCALE_GLOBE ) {
    r = 1.75;
  } 
  else if ( projectionScale == SCALE_NORTHWEST ) {
    r = 2.25;
  } 
  else if ( projectionScale == SCALE_LOWER48 ) {
    r = 3.0;
  }
  else { // SCALE_FREE
    /* change value of r based on current scaleFactor */
    r = map(scaleFactor, 1.0, 8.0, 1.75, 3.50);
  } 

  /* drawMapLayer renders the borders, coasts, graticules, etc. */
  drawMapLayer();

  /* render tweets to map */
  if ( showTweets ) {
    pushStyle();
    if ( !cumulativeMode ) {
      noStroke();
      float age;
      // loop over all tweets
      for (int i = 0; i < n; i++ ) {
        // did tweet happen before current time...
        // but after the beginning of the window we're displaying?
        if ( times[i] < t && times[i] > t-windowSize ) {
          // make dot fade more the older it is
          //   newest is maroon (128,0,0,255)
          //   oldest is (transparent) "eggplant" rgb(128,0,96,0);
          age = map(t-times[i], 0, windowSize, 5, -5 );
          age = 1.0 / ( 1.0 + exp(age) );
          fill(128, 0, 32+map(t-times[i], 0, windowSize, 0, 96), map(age, 0, 1, 255, 0));
          ellipse(calcX(lat[i], lon[i]), calcY(lat[i], lon[i]), r, r);
        }
      }
    } 
    else { // using cumulative mode
      noFill();
      stroke(71, 61, 140, 64); // "Dark Slate Blue" (bluish purple)
      strokeWeight(r);
      // loop over all tweets
      for (int i = 0; i < n; i++ ) {
        // did tween happen before current time?
        if ( times[i] < t ) {
          point(calcX(lat[i], lon[i]), calcY(lat[i], lon[i]));
        }
      }
    }
    popStyle();
  }

  /* draw donations */
  if ( showDonations ) {
    pushStyle();
    noStroke();
    double tt = 0;
    if ( !cumulativeMode ) {
      noStroke();
      float age;
      for (int i = 0; i < donations.length; i++ ) {
        tt = donations[i].getTime();
        if ( tt < t && tt > t - windowSize ) {
          age = map((float)(t-tt), 0, windowSize, 5, -5 );
          age = 1.0 / ( 1.0 + exp(age) ); //sigmoidal decay
          fill(0, 128, 16, map(age, 0, 1, 255, 0));
          ellipse(donations[i].getX(), donations[i].getY(), r+1, r+1);
        }
      }
    } 
    else { // using cumulative mode
      strokeWeight(r+0.5);
      stroke(0, 142, 16, 64);
      for (int i = 0; i < donations.length; i++ ) {
        if ( donations[i].getTime() < t ) {
          point(donations[i].getX(), donations[i].getY());
        }
      }
    }
    popStyle();
  }

  /* highlight location of hurricane */
  if ( showStormEye ) {
    pushStyle();
    if ( !cumulativeMode ) {
      float age;
      float x, y;
      for (int i = 0; i < pathTimes.length; i++ ) {
        if ( pathTimes[i] < t && pathTimes[i] > t - windowSize ) {
          x = calcX( pathLat[i], pathLon[i] );
          y = calcY( pathLat[i], pathLon[i] );
          strokeWeight(2);
          stroke(160, 8, 16, map(t-pathTimes[i], 0, windowSize, 255, 0));  // "carmine"
          fill(255, 255, 255, map(t-pathTimes[i], 0, windowSize, 255, 0)); // white
          ellipse(x, y, 2*r, 2*r);
          // fade maroon to eggplant based on age
          age = map(t-pathTimes[i], 0, windowSize, 5, -5 );
          age = 1.0 / ( 1.0 + exp(age) );
          fill(128, 0, 32+map(t-pathTimes[i], 0, windowSize, 0, 96), map(age, 0, 1, 255, 0));
          noStroke();
          ellipse(x, y, r, r);
        }
      }
    } 
    else { // using cumulative mode
      strokeWeight(1.5*r);
      stroke(0, 0, 255);
      noFill();
      for (int i = 0; i < pathTimes.length; i++ ) {
        if ( pathTimes[i] < t ) {
          point( calcX(pathLat[i], pathLon[i]), calcY(pathLat[i], pathLon[i]) );
        }
      }
    }
    popStyle();
  }

  /* step forward in time */
  t += stepSize;

  /* display current time & date in upper left
   (I've had to do some weird contortions here to make this work.
   Watch out for floating point precision errors if you
   fiddle with this.) */
  fill(30, 30, 50);
  textAlign(LEFT);
  double tempTime = t+minTime;
  Date currTime = new java.util.Date((long)(tempTime*1000.0));
  drawSun(currTime);
  String date = new java.text.SimpleDateFormat("HH:mm:ss yyyy-MM-dd").format(currTime);
  text(date, 35, 25);

  /* display dataset name in lower left */
  text(datasetDisplayName, 20, height-15);
  textAlign(RIGHT);

  /* save each frame so it can be rendered to video later */
  if ( saveFrames ) {
    saveFrame(outputName+"/frame-####.jpg");
  }

  /* end of animation */
  if ( t > duration ) {
    save(outputName+".tiff");
    exit();
  }

  /* NB: everything drawn after this point won't show up in the
   output images so this where you want to display anything
   you're using for debugging */

  /* display framerate in bottom right corner */
  text(nf(frameRate, 2, 1), width-20, height-15);

  /* display name of map projection in upper right */
  text(projName, width-20, 25);

  /* display zoom level in bottom right corner */
  text(scaleFactor, width-20, height-50);

  /* progress indiactor along bottom of screen */
  pushStyle();
  strokeWeight(3);
  stroke(255);
  line(0, height, width, height);
  stroke(0);
  line(0, height, map((float)(t), 0, (float)(duration), 0, width), height);
  popStyle();
}

